// Package genswagger contains method for generate swagger config from protobuf file.
package genswagger

import (
	"encoding/json"
	"fmt"
	"strings"

	"gitlab.com/pjrpc/pjrpc/cmd/protoc-gen-go-pjrpc/genpjrpc"
	"gitlab.com/pjrpc/pjrpc/cmd/protoc-gen-go-pjson/genpjson"
	"google.golang.org/protobuf/compiler/protogen"
	"google.golang.org/protobuf/reflect/protoreflect"

	"gitlab.com/pjrpc/proto-swagger-pjrpc/model"
)

const methodsTagName = "methods"

type generator struct {
	gen        *protogen.Plugin
	file       *protogen.GeneratedFile
	info       *filesInfo
	service    *serviceInfo
	spec       *model.Swagger
	allSchemas map[string]*model.Schema
}

func (g *generator) getTypeName(goIdent protogen.GoIdent) string {
	return g.file.QualifiedGoIdent(goIdent)
}

// genInfo generates default info texts.
func (g *generator) genInfo() {
	g.spec.Swagger = swaggerVersion

	if g.spec.Info == nil {
		g.spec.Info = new(model.Info)
	}

	setStringIfEmpty(&g.spec.Info.Title, g.service.Name)

	setStringIfEmpty(&g.spec.Info.Description, jsonComment(g.service.Service.Comments).String())

	if !HideVersions {
		if g.spec.Info.Description != "" {
			g.spec.Info.Description += "\n"
		}

		g.spec.Info.Description = addText(
			g.spec.Info.Description,
			fmt.Sprintf("\nGenerated by proto-swagger-pjrpc.\n- protoc-gen-swagger-pjrpc: %s;\n", Version),
		)
	}

	setStringIfEmpty(&g.spec.Info.Version, ServicesVersion)

	setStringIfEmpty(&g.spec.BasePath, "/")

	g.spec.Produces = []string{"application/json"}
	g.spec.Consumes = []string{"application/json"}
}

// genEnumSchema converts proto enum to swagger schema.
func (g *generator) genEnumSchema(enum *genpjson.EnumInfo) {
	typeName := g.getTypeName(enum.GoIdent)

	schema := &model.Schema{
		Name:        typeName,
		Description: addDeprecated(enum.Desc.Options(), enum.Comments),
		Type:        model.SchemaTypeString,
		Enum:        make([]string, len(enum.Values)),
	}

	for i, value := range enum.Values {
		schema.Enum[i] = string(value.Desc.Name())
		schema.Description += fmt.Sprintf("\n * `%s` - %s",
			value.Desc.Name(),
			addDeprecated(value.Desc.Options(), value.Comments),
		)
	}

	g.allSchemas[typeName] = schema
}

// genFieldKind converts primitive protobuf field kind to swagger schema.
func (g *generator) genFieldKind(field *protogen.Field) (*model.Schema, error) {
	if field.Desc.IsMap() { // this method can't to converts map field
		return nil, nil
	}

	schema := new(model.Schema)

	switch field.Desc.Kind() {
	case protoreflect.BoolKind:
		schema.Type = model.SchemaTypeBoolean

	case protoreflect.FloatKind:
		schema.Type = model.SchemaTypeNumber
		schema.Format = model.SchemaFormatFloat

	case protoreflect.DoubleKind:
		schema.Type = model.SchemaTypeNumber
		schema.Format = model.SchemaFormatDouble

	case protoreflect.Int32Kind, protoreflect.Sint32Kind, protoreflect.Sfixed32Kind:
		schema.Type = model.SchemaTypeInteger
		schema.Format = model.SchemaFormatInt32

	case protoreflect.Uint32Kind, protoreflect.Fixed32Kind:
		schema.Type = model.SchemaTypeInteger
		schema.Format = model.SchemaFormat("uint32")

	case protoreflect.Int64Kind, protoreflect.Sint64Kind, protoreflect.Sfixed64Kind:
		schema.Type = model.SchemaTypeInteger
		schema.Format = model.SchemaFormatInt64

	case protoreflect.Uint64Kind, protoreflect.Fixed64Kind:
		schema.Type = model.SchemaTypeInteger
		schema.Format = model.SchemaFormat("uint64")

	case protoreflect.StringKind:
		schema.Type = model.SchemaTypeString

	case protoreflect.BytesKind:
		schema.Type = model.SchemaTypeString
		schema.Format = model.SchemaFormatByte

	case protoreflect.EnumKind:
		schema = g.allSchemas[g.getTypeName(field.Enum.GoIdent)].Reference()

	case protoreflect.MessageKind, protoreflect.GroupKind:
		msgSchema, err := g.genMessageSchema(g.info.allMessages[field.Message.GoIdent.String()])
		if err != nil {
			return nil, fmt.Errorf("genMessageSchema: %w", err)
		}
		schema = msgSchema.Reference()
	}

	return schema, nil
}

// genMapSchema converts protobuf map to swagger schema.
func (g *generator) genMapSchema(field *genpjson.FieldInfo) (*model.Schema, error) {
	key, err := g.genFieldKind(field.Message.Fields[0])
	if err != nil {
		return nil, fmt.Errorf("genFieldKind map key: %w", err)
	}

	value, err := g.genFieldKind(field.Message.Fields[1])
	if err != nil {
		return nil, fmt.Errorf("genFieldKind map value: %w", err)
	}

	var keyType string
	switch key.Type { // nolint:exhaustive // only these types are checked.
	case model.SchemaTypeInteger:
		keyType = fmt.Sprintf("%s($%s)", key.Type, key.Format)
	case model.SchemaTypeString:
		keyType = string(key.Type)
	default:
		return nil, fmt.Errorf("%w: key type: '%s'", ErrInvalidKeyType, key.Type)
	}

	valueType := string(model.SchemaTypeObject)
	if value.Type != "" {
		valueType = string(value.Type)
		if value.Format != "" {
			valueType = fmt.Sprintf("%s($%s)", valueType, value.Format)
		}
	}

	schema := &model.Schema{
		Type:  model.SchemaTypeObject,
		Title: fmt.Sprintf("map[%s]%s", keyType, valueType),

		AdditionalProperties: value,
	}

	return schema, nil
}

// genJSONType converts any field type to swagger schema.
func (g *generator) genJSONType(field *genpjson.FieldInfo) (*model.Schema, error) {
	kind, err := g.genFieldKind(field.Field)
	if err != nil {
		return nil, fmt.Errorf("genFieldKind: %w", err)
	}

	var schema *model.Schema

	switch {
	case field.Desc.IsList():
		schema = &model.Schema{
			Type:  model.SchemaTypeArray,
			Items: kind,
		}

	case field.Desc.IsMap():
		schema, err = g.genMapSchema(field)
		if err != nil {
			return nil, fmt.Errorf("genMapSchema: %w", err)
		}

	default:
		schema = kind
	}

	schema.Description = addDeprecated(field.Desc.Options(), field.Comments)

	return schema, nil
}

// genMessageSchema converts proto messages to swagger schemas.
// returns new schema or already generated.
func (g *generator) genMessageSchema(msg *genpjson.MessageInfo) (*model.Schema, error) {
	if msg.Desc.IsMapEntry() {
		return nil, nil
	}

	typeName := g.getTypeName(msg.GoIdent)
	if schema, ok := g.allSchemas[typeName]; ok {
		return schema, nil
	}

	schema := &model.Schema{
		Name:        typeName,
		Description: addDeprecated(msg.Desc.Options(), msg.Comments),
		Type:        model.SchemaTypeObject,
		Properties:  make(model.OrderedMap, 0, len(msg.FieldsInfo)),
	}

	g.allSchemas[typeName] = schema

	for _, field := range msg.FieldsInfo {
		fieldProp, err := g.genJSONType(field)
		if err != nil {
			return nil, fmt.Errorf("genJSONType (%s): %w", field.JSONName, err)
		}

		schema.Properties = schema.Properties.Add(field.JSONName, fieldProp)
	}

	return schema, nil
}

// genSchemas converts all proto messages to swagger schemas.
func (g *generator) genSchemas() error {
	g.allSchemas = make(map[string]*model.Schema, len(g.info.allMessages)+len(g.info.allEnums))

	for _, enum := range g.info.allEnums {
		g.genEnumSchema(enum)
	}

	for _, msg := range g.info.allMessages {
		_, err := g.genMessageSchema(msg)
		if err != nil {
			return fmt.Errorf("genMessageSchema: %w", err)
		}
	}

	return nil
}

func (g *generator) copySchemaToDefinitions(goIdent protogen.GoIdent) *model.Schema {
	typeName := g.getTypeName(goIdent)

	schema, ok := g.spec.Definitions[typeName]
	if ok {
		return schema
	}

	schema = g.allSchemas[typeName]
	g.spec.Definitions[typeName] = schema

	return schema
}

// genDefinitions copy schema with all sub types from allSchemas to spec Definitions.
func (g *generator) genDefinitions(msg *protogen.Message) *model.Schema {
	schema := g.copySchemaToDefinitions(msg.GoIdent)

	// finds sub types of the message
	for _, field := range msg.Fields {
		if field.Desc.IsMap() {
			continue
		}

		switch field.Desc.Kind() { // nolint:exhaustive // only these types are needed.
		case protoreflect.EnumKind:
			g.copySchemaToDefinitions(field.Enum.GoIdent)

		case protoreflect.MessageKind, protoreflect.GroupKind:
			g.genDefinitions(field.Message)
		}
	}

	return schema
}

func (*generator) genJSONRPCInputSchema(methodName string, input *model.Schema) *model.Schema {
	props := make(model.OrderedMap, 0, 4) // nolint:gomnd // 4 fields of the schema

	props = props.
		Add("jsonrpc", &model.Schema{
			Description: `A String specifying the version of the JSON-RPC protocol. MUST be exactly "2.0".`,
			Type:        model.SchemaTypeString,
			Enum:        []string{"2.0"},
		}).
		Add("id", &model.Schema{
			Description: `An identifier established by the Client.`,
			Type:        model.SchemaTypeString,
			Format:      model.SchemaFormat("uuid"),
		}).
		Add("method", &model.Schema{
			Description: `A String containing the name of the method to be invoked.`,
			Type:        model.SchemaTypeString,
			Enum:        []string{methodName},
		}).
		Add("params", input.Reference())

	return &model.Schema{
		Type:       model.SchemaTypeObject,
		Required:   []string{"jsonrpc", "id", "method"},
		Properties: props,
	}
}

func (*generator) genJSONRPCOutputSchema(output *model.Schema) *model.Schema {
	props := make(model.OrderedMap, 0, 4) // nolint:gomnd // 4 fields of the schema

	props = props.
		Add("jsonrpc", &model.Schema{
			Description: `A String specifying the version of the JSON-RPC protocol. MUST be exactly "2.0".`,
			Type:        model.SchemaTypeString,
			Enum:        []string{"2.0"},
		}).
		Add("id", &model.Schema{
			Description: `It MUST be the same as the value of the id member in the Request.`,
			Type:        model.SchemaTypeString,
			Format:      model.SchemaFormat("uuid"),
		}).
		Add("error", &model.Schema{
			Description: `REQUIRED on error. This member MUST NOT exist if there was no error triggered during invocation.`,
			Type:        model.SchemaTypeObject,
			Properties: make(model.OrderedMap, 0, 4).
				Add("code", &model.Schema{
					Description: `A Number that indicates the error type that occurred.`,
					Type:        model.SchemaTypeInteger,
					Format:      model.SchemaFormatInt64,
				}).
				Add("message", &model.Schema{
					Description: `A String providing a short description of the error.`,
					Type:        model.SchemaTypeString,
				}).
				Add("data", &model.Schema{
					Description: `A Primitive or Structured value that contains additional information about the error.`,
					Type:        model.SchemaTypeObject,
				}),
		}).
		Add("result", output.Reference())

	return &model.Schema{
		Type:       model.SchemaTypeObject,
		Required:   []string{"jsonrpc", "id"},
		Properties: props,
	}
}

func (*generator) getMethodTags(comments protogen.CommentSet) []string {
	lines := getLeadingComments(comments)

	for _, line := range lines {
		if !strings.Contains(line, swaggerCommentTag) {
			continue
		}

		line = strings.TrimSpace(line)
		line = strings.TrimPrefix(line, swaggerCommentTag+":")

		tags := strings.Split(line, ",")
		for i := range tags {
			tags[i] = strings.TrimSpace(tags[i])
		}

		return tags
	}

	return []string{methodsTagName}
}

// genMethodSummary return first line of the leading comments of method.
func (*generator) genMethodSummary(method *genpjrpc.MethodInfo) string {
	lines := getLeadingComments(method.Method.Comments)

	summary := strings.Replace(lines[0], method.Method.GoName, "", 1) // removes method name from comment.

	return strings.TrimSpace(summary)
}

func (g *generator) genMethod(method *genpjrpc.MethodInfo, input, output *model.Schema) *model.PathItem {
	op := &model.Operation{
		Tags:        g.getMethodTags(method.Method.Comments),
		Summary:     g.genMethodSummary(method),
		Description: jsonComment(method.Method.Comments).String(),
		Deprecated:  isDeprecated(method.Method.Desc.Options()) || isDeprecated(g.service.Service.Desc.Options()),
		OperationID: method.RPCName,
		Parameters: []*model.Parameter{{
			Name:        "body",
			In:          model.ParameterInBody,
			Description: input.Description,
			Required:    true,
			Schema:      g.genJSONRPCInputSchema(method.RPCName, input),
		}},
		Responses: map[string]*model.Response{
			"200": {
				Description: output.Description,
				Schema:      g.genJSONRPCOutputSchema(output),
			},
		},
	}

	return &model.PathItem{
		Post: op,
	}
}

// genMethods converts proto methods to swagger.
// Returns list of the unique tags of the methods.
func (g *generator) genMethods() []*model.Tag {
	g.spec.Paths = make(model.OrderedMap, 0, len(g.service.Methods))

	tags := make([]string, 0, len(g.service.Methods))

	for _, method := range g.service.Methods {
		input := g.genDefinitions(method.Method.Input)
		output := g.genDefinitions(method.Method.Output)

		path := g.genMethod(method, input, output)
		tags = append(tags, path.Post.Tags...)

		g.spec.Paths = g.spec.Paths.Add("/#"+method.RPCName, path)
	}

	tags = uniqueStrings(tags)

	res := make([]*model.Tag, len(tags))
	for i, tag := range tags {
		res[i] = &model.Tag{
			Name: tag,
		}
	}

	return res
}

// genService fills the swagger spec and puts it in a file.
func (g *generator) genService() error {
	g.genInfo()

	err := g.genSchemas()
	if err != nil {
		return fmt.Errorf("genSchemas: %w", err)
	}

	if AllTypes {
		g.spec.Definitions = make(map[string]*model.Schema, len(g.allSchemas))
		for typeName, schema := range g.allSchemas {
			g.spec.Definitions[typeName] = schema
		}
	} else {
		// nolint:gomnd // 2 is one input and one output per method.
		g.spec.Definitions = make(map[string]*model.Schema, len(g.service.Methods)*2)
	}

	tags := g.genMethods()

	g.spec.Tags = append(g.spec.Tags, tags...)

	err = json.NewEncoder(g.file).Encode(g.spec)
	if err != nil {
		return fmt.Errorf("failed encode spec: %w", err)
	}

	return nil
}

func genServices(gen *protogen.Plugin, services []string, info *filesInfo, descriptions model.ServiceDescriptions) error {
	for _, serviceName := range services {
		service, ok := info.services[serviceName]
		if !ok {
			return fmt.Errorf("%w: want '%s', have: {%s}",
				ErrServiceNotFound, serviceName, strings.Join(services, ", "))
		}

		spec, ok := descriptions[serviceName]
		if !ok {
			spec = new(model.Swagger)
		}

		g := &generator{
			gen:     gen,
			info:    info,
			service: service,
			spec:    spec,
		}

		fileName := "swagger.json"
		if len(services) != 1 {
			fileName = fmt.Sprintf("swagger_%s.json", g.service.Name)
		}

		g.file = g.gen.NewGeneratedFile(fileName, g.service.file.GoImportPath)

		err := g.genService()
		if err != nil {
			return fmt.Errorf("genService: %w", err)
		}
	}

	return nil
}

// Run loops through files to generate and creates spec file.
func Run(gen *protogen.Plugin) error {
	genpjrpc.MethodsWithServiceName = MethodsWithServiceName
	setStringIfEmpty(&ServicesVersion, "v0.0.0-unknown")

	descriptions, err := loadServiceDescriptions(ServiceDescriptionsPath)
	if err != nil {
		return fmt.Errorf("loadServiceDescriptions: %w", err)
	}

	info := getFilesInfo(gen, gen.Files)

	serviceNames := make([]string, 0, len(info.services))
	for name := range info.services {
		serviceNames = append(serviceNames, name)
	}

	if len(serviceNames) == 0 {
		return ErrNoServicesInFiles
	}

	if len(GenerateServices) == 0 {
		GenerateServices = serviceNames
	}

	err = genServices(gen, GenerateServices, info, descriptions)
	if err != nil {
		return fmt.Errorf("genServices: %w", err)
	}

	return nil
}
